(* Coq version 8.5 *)

(* BEGINNING OF THEOREM *)

Variable Label : Set.

(* We assume the set of labels is finite. *)
Variable all_labels : list Label.

(* To formalize the finiteness of Label, we need something like the following. *)
(*
Inductive Foreach (A : Type) (P : A->Type) : list A -> Type :=
  | Foreach_nil  : Foreach A P nil
  | Foreach_cons : forall x xs, P x -> Foreach A P xs -> Foreach A P (cons x xs)
  .
Variable finitely_many_labels : forall P, Foreach Label P all_labels -> forall k, P k.
*)

Variable Char : Set.
Variable Name : Set.
Variable Address : Set.
Variable Handle : Set.

Inductive Term : Set :=
  | TVar : Name -> Term
  | TLam : Name -> Term -> Term
  | TApp : Term -> Term -> Term
  | TChar : Char -> Term
  | TPublic : Term -> Term
  | TFaceted : Label -> Term -> Term -> Term
  | TBottom : Term
  | TRetFac : Term -> Term
  | TBindFac : Term -> Term -> Term
  | TRetFIO : Term -> Term
  | TBindFIO : Term -> Term -> Term
  | TProd : Term -> Term
  | TNew : Term -> Term
  | TRead : Term -> Term
  | TWrite : Term -> Term -> Term
  | TGet : Handle -> Term
  | TPut : Handle -> Term -> Term
  | TUnit : Term
  | TAddress : Address -> Term
  .

Inductive IsValue : Term -> Set :=
  | VPublic : forall t, IsValue (TPublic t)
  | VFaceted : forall k t1 t2, IsValue t1 -> IsValue t2 -> IsValue (TFaceted k t1 t2)
  | VBottom : IsValue TBottom
  | VRetFIO : forall t, IsValue (TRetFIO t)
  | VBindFIO : forall t1 t2, IsValue (TBindFIO t1 t2)
  | VProd : forall t, IsValue t -> IsValue (TProd t)
  | VNew : forall t, IsValue (TNew t)
  | VRead : forall t, IsValue (TRead t)
  | VWrite : forall h t, IsValue (TWrite h t)
  | VGet : forall h, IsValue (TGet h)
  | VPut : forall h t, IsValue (TPut h t)
  | VLam : forall x t, IsValue (TLam x t)
  | VChar : forall ch, IsValue (TChar ch)
  | VAddress : forall a, IsValue (TAddress a)
  | VUnit : IsValue TUnit
  .
Definition NotAValue t := IsValue t -> False.

Inductive EvalContext :=
  | EApp : Term -> EvalContext
  | EBindFac : Term -> EvalContext
  | EFaceted1 : Label -> Term -> EvalContext
  | EFaceted2 : Label -> forall t, IsValue t -> EvalContext
  | EProd : EvalContext
  .

Definition String : Set := list Char.
Definition Memory : Set := Address -> Term.
Definition FileSystem : Set := Handle -> String.
Definition Store : Set := Memory * FileSystem.

Fixpoint fill (E : EvalContext) (t : Term) : Term :=
  match E with
  | EApp t2 => TApp t t2
  | EBindFac t2 => TBindFac t t2
  | EFaceted1 k t2 => TFaceted k t t2
  | EFaceted2 k t1 _ => TFaceted k t1 t
  | EProd => TProd t
  end.

Variable subs : Term -> Name -> Term -> Term.

Inductive Eval : Term -> Term -> Set :=
  | E_Val : forall t,
      IsValue t ->
      Eval t t
  | E_App : forall t x t1 v,
      Eval (subs t x t1) v ->
      Eval (TApp (TLam x t) t1) v
  | E_Ctxt : forall t v1 E v2,
      NotAValue t ->
      Eval t v1 ->
      Eval (fill E v1) v2 ->
      Eval (fill E t) v2
  | E_Ret : forall t,
      Eval (TRetFac t) (TPublic t)
  | E_Bind_P : forall t2 t1 F,
      Eval (TApp t2 t1) F ->
      Eval (TBindFac (TPublic t1) t2) F
  | E_Bind_F : forall F1 (t3:Term) F1' F2 t3 F2' k,
      Eval (TBindFac F1 t3) F1' ->
      Eval (TBindFac F2 t3) F2' ->
      IsValue F1 ->
      IsValue F2 ->
      Eval (TBindFac (TFaceted k F1 F2) t3) (TFaceted k F1' F2')
  | E_Bind_B : forall t,
      Eval (TBindFac TBottom t) TBottom
  .

Definition PC := Label -> option bool.
Variable eq_dec_label : forall (k1 k2 : Label), {k1=k2}+{k1<>k2}.
Definition add (pc:PC) k1 k2 :=
  if eq_dec_label k1 k2 then
    Some true
  else
    pc k2
  .
Definition subtract (pc:PC) k1 k2 :=
  if eq_dec_label k1 k2 then
    Some false
  else
    pc k2
  .

Definition View := Label -> bool.
Variable view : Handle -> View.
Definition Visible (pc:PC) L :=
  forall k,
       (pc k = Some true  -> L k = true )
    /\ (pc k = Some false -> L k = false).
Definition NotVisible pc L := not (Visible pc L).
Definition view2pc (L:View) k := Some (L k).

Variable eq_dec_addr : forall a1 a2 : Address, {a1=a2}+{a1<>a2}.
Definition read (s:Store) a := fst s a.
Definition overwrite (s:Store) a t :=
  let (m, fs) := s in
  let m' := fun a' =>
    if eq_dec_addr a a' then
      t
    else
      m a'  in
  (m', fs).
Variable eq_dec_handle : forall h1 h2 : Handle, {h1=h2}+{h1<>h2}.
Definition contents (s:Store) h := snd s h.
Definition replace (s:Store) h chs :=
  let (m, fs) := s in
  let fs' := fun h' =>
    if eq_dec_handle h h' then
      chs
    else
      fs h'  in
  (m, fs').
Definition NotInDomain a s := read s a = TBottom.

Definition facet : PC -> Term -> Term -> Term :=
  fun pc t1 t2 =>
    let g := fix g ks : Term :=
      match ks with
      | nil => t1
      | cons k ks =>
          match pc k with
          | None       => g ks
          | Some true  => TFaceted k (g ks) t2
          | Some false => TFaceted k t2 (g ks)
          end
      end                                       in
  g all_labels.

Inductive Run : PC -> Store -> Term -> Store -> Term -> Prop :=
  | F_Ret : forall s t pc,
      Run pc s (TRetFIO t) s t
  | F_Bind : forall t1 t2 t3 t4 A1 A2 s0 s1 s2 pc,
      Eval t1 A1 ->
      Run pc s0 A1 s1 t3 ->
      Eval (TApp t2 t3) A2 ->
      Run pc s1 A2 s2 t4 ->
      Run pc s0 (TBindFIO t1 t2) s2 t4
  | F_Prod_P : forall t A s s' t' pc,
      Eval t A ->
      Run pc s A s' t' ->
      Run pc s (TProd (TPublic t)) s' t'
  | F_Prod_F1 : forall k pc s F1 s' t' F2,
      pc k = Some true ->
      Run pc s (TProd F1) s' t' ->
      Run pc s (TProd (TFaceted k F1 F2)) s' t'
  | F_Prod_F2 : forall k pc s F1 s' t' F2,
      pc k = Some false ->
      Run pc s (TProd F2) s' t' ->
      Run pc s (TProd (TFaceted k F1 F2)) s' t'
  | F_Prod_F3 : forall k pc s0 s1 s2 t1 t2 F1 F2,
      pc k = None ->
      Run (add pc k) s0 (TProd F1) s1 t1 ->
      Run (subtract pc k) s1 (TProd F2) s2 t2 ->
      Run pc s0 (TProd (TFaceted k F1 F2)) s2 (TFaceted k t1 t2)
  | F_New : forall t F a s pc,
      Eval t F -> (* Is strictness needed here? *)
      NotInDomain a s ->
      Run pc s (TNew t) (overwrite s a (facet pc F TBottom)) (TAddress a)
  | F_Read : forall t a s pc,
      Eval t (TAddress a) ->
      Run pc s (TRead t) s (read s a)
  | F_Write : forall t1 a s pc t2,
      Eval t1 (TAddress a) ->
      Run pc s (TWrite t1 t2) (overwrite s a (facet pc t2 (read s a))) (TPublic TUnit)
  | F_Get : forall h pc str ch s,
      Visible pc (view h) ->
      contents s h = cons ch str ->
      Run pc s (TGet h) (replace s h str) (facet (view2pc (view h)) (TPublic (TChar ch)) TBottom)
  | F_Get_2 : forall h pc s,
      NotVisible pc (view h) ->
      Run pc s (TGet h) s TBottom
  | F_Put : forall t h pc ch s,
      Visible pc (view h) ->
      Eval t (TChar ch) ->
      Run pc s (TPut h t) (replace s h (app (contents s h) (ch :: nil))) TUnit
  | F_Put_2 : forall t h pc s,
      NotVisible pc (view h) ->
      Run pc s (TPut h t) s TUnit
  .

Definition lift f : Term -> Term :=
  fix g t :=
    match t with
    | TAddress a => TAddress (f a)

    | TFaceted k t1 t2 => TFaceted k (g t1) (g t2)
    | TLam x t => TLam x (g t)
    | TApp t1 t2 => TApp (g t1) (g t2)
    | TPublic t => TPublic (g t)
    | TRetFac t => TRetFac (g t)
    | TBindFac t1 t2 => TBindFac (g t1) (g t2)
    | TRetFIO t => TRetFIO (g t)
    | TBindFIO t1 t2 => TBindFIO (g t1) (g t2)
    | TProd t => TProd (g t)
    | TNew t => TNew (g t)
    | TRead t => TRead (g t)
    | TWrite t1 t2 => TWrite (g t1) (g t2)
    | TPut h t => TPut h (g t)
    | t => t
    end.

Definition EQUIV s1 t1 s2 t2 :=
  exists f, t1 = lift f t2 /\ (forall a, read s1 a = lift f (read s2 (f a)))
                           /\ (forall h, contents s1 h = contents s2 h)
  .

Variable eq_dec_view : forall (L1 L2 : View), {L1=L2}+{L1<>L2}.

Fixpoint pt (L:View) t : Term :=
    match t with
    | TFaceted k t1 t2 =>
        if L k then
          pt L t1
        else
          pt L t2
    | TLam x t => TLam x (pt L t)
    | TApp t1 t2 => TApp (pt L t1) (pt L t2)
    | TPublic t => TPublic (pt L t)
    | TRetFac t => TRetFac (pt L t)
    | TBindFac t1 t2 => TBindFac (pt L t1) (pt L t2)
    | TRetFIO t => TRetFIO (pt L t)
    | TBindFIO t1 t2 => TBindFIO (pt L t1) (pt L t2)
    | TProd t => TProd (pt L t)
    | TNew t => TNew (pt L t)
    | TRead t => TRead (pt L t)
    | TWrite t1 t2 => TWrite (pt L t1) (pt L t2)
    | TGet h =>
        if eq_dec_view L (view h) then
          TGet h
        else
          TRetFIO TBottom
    | TPut h t =>
        if eq_dec_view L (view h) then
          TPut h (pt L t)
        else
          TRetFIO TUnit
    | t => t
    end.

Definition ps (L:View) (s:Store) : Store :=
  let m a := pt L (read s a)  in
  let fs h :=
    if eq_dec_view L (view h) then
      contents s h
    else
      nil  in
  (m, fs).

Definition empty : PC := fun _ => None.

Definition NONINTERFERENCE :=
    forall L sigma1 sigma2 A1 A2 sigma1' t1 sigma2' t2,
      EQUIV (ps L sigma1) (pt L A1) (ps L sigma2) (pt L A2) ->
      Run empty sigma1 A1 sigma1' t1 ->
      Run empty sigma2 A2 sigma2' t2 ->
      EQUIV (ps L sigma1') (pt L t1) (ps L sigma2') (pt L t2).

(* END OF THEOREM *)
(* ------------------------------------------------------------------ *)
(* BEGINNING OF PROOF *)

Inductive Run' : Store -> Term -> Store -> Term -> Prop :=
  | G_Ret : forall s t,
      Run' s (TRetFIO t) s t
  | G_Bind : forall t1 t2 t3 t4 A1 A2 s0 s1 s2,
      Eval t1 A1 ->
      Run' s0 A1 s1 t3 ->
      Eval (TApp t2 t3) A2 ->
      Run' s1 A2 s2 t4 ->
      Run' s0 (TBindFIO t1 t2) s2 t4
  | G_Prod_P : forall t A s s' t',
      Eval t A ->
      Run' s A s' t' ->
      Run' s (TProd (TPublic t)) s' t'
  | G_New : forall t F s a,
      Eval t F ->
      NotInDomain a s ->
      Run' s (TNew t) (overwrite s a F) (TAddress a)
  | G_Read : forall t a s,
      Eval t (TAddress a) ->
      Run' s (TRead t) s (read s a)
  | G_Write : forall t1 a s t2,
      Eval t1 (TAddress a) ->
      Run' s (TWrite t1 t2) (overwrite s a t2) (TPublic TUnit)
  | G_Get : forall h str ch s,
      contents s h = cons ch str ->
      Run' s (TGet h) (replace s h str) (TPublic (TChar ch))
  | G_Put : forall t h ch s,
      Eval t (TChar ch) ->
      Run' s (TPut h t) (replace s h (app (contents s h) (ch :: nil))) TUnit
  .

Lemma Z1 : forall t L, IsValue t -> IsValue (pt L t).
 induction 1.
               apply VPublic.
              simpl; destruct (L k); auto.
             apply VBottom.
            apply VRetFIO.
           apply VBindFIO.
          apply VProd; auto.
         apply VNew.
        apply VRead.
       apply VWrite.
      simpl; destruct (eq_dec_view L (view h)); [
          apply VGet
        | apply VRetFIO
      ].
     simpl; destruct (eq_dec_view L (view h)); [
         apply VPut
       | apply VRetFIO
     ].
    apply VLam.
   apply VChar.
  apply VAddress.
 apply VUnit.
Qed.

Variable Z2 : forall t1 x t2 L, pt L (subs t1 x t2) = subs (pt L t1) x (pt L t2).

Definition Z3_helper E :=
  match E with
  | EFaceted1 _ _ => False
  | EFaceted2 _ _ _ => False
  | _ => True
  end.
Lemma Z3 : forall L E, Z3_helper E -> {E' | forall t, pt L (fill E t) = fill E' (pt L t)}.
 induction E; intros.
     exists (EApp (pt L t)); auto.
    exists (EBindFac (pt L t)); auto.
   inversion H.
  inversion H.
 exists EProd; auto.
Qed.

Lemma eval_deterministic : forall t v1, Eval t v1 -> forall v2, Eval t v2 -> v1 = v2.
Admitted.

Lemma double_eval : forall t v1 v2, Eval t v1 -> Eval v1 v2 -> v1 = v2.
 intros.
 assert (IsValue v1).
  clear H0.
  induction H; auto.
    apply VPublic.
   apply VFaceted; auto.
  apply VBottom.
 assert (Eval v1 v1); try apply E_Val; auto.
 apply eval_deterministic with v1; auto.
Qed.

Lemma eval_val : forall v1 v2, IsValue v1 -> Eval v1 v2 -> v1 = v2.
 intros.
 apply eval_deterministic with v1; auto.
 apply E_Val; auto.
Qed.

Lemma isvalue_dec : forall t, IsValue t + NotAValue t.
 set (helper := (VPublic, VLam, VChar, VFaceted, VBottom, VRetFIO, VBindFIO, VProd, VNew, VWrite, VRead, VGet, VPut, VUnit, VAddress)).
 intros.
 induction t; try (intuition; fail).
 all: try (right; intro; inversion H; fail).
  destruct IHt1, IHt2; [left; intuition | | |]; right; intro; inversion H; contradiction.
 destruct IHt; intuition.
 right; intro; inversion H; contradiction.
Qed.

Lemma ZZ : forall t, IsValue t + NotAValue t.
 set (helper := (VPublic, VLam, VChar, VFaceted, VBottom, VRetFIO, VBindFIO, VProd, VNew, VWrite, VRead, VGet, VPut, VUnit, VAddress)).
 intros.
 induction t; try (intuition; fail).
 all: try (right; intro; inversion H; fail).
  destruct IHt1, IHt2; [left; intuition | | |]; right; intro; inversion H; contradiction.
 destruct IHt; intuition.
 right; intro; inversion H; contradiction.
Qed.

Theorem pure_projection :
    forall t v L,
      Eval t v ->
      Eval (pt L t) (pt L v).
 induction 1.
       apply E_Val.
       apply Z1.
       assumption.
      rewrite Z2 in IHEval.
      apply E_App.
      assumption.
     destruct E.
         destruct (isvalue_dec (pt L t)).
          destruct (Z3 L (EApp t0)) as (E', H1); try exact I.
          rewrite H1.
          rewrite H1 in IHEval2.
          rewrite eval_val with (v1 := pt L t) (v2 := pt L v1); auto.
         apply E_Ctxt with (E := EApp (pt L t0)) (v1 := pt L v1) (t := pt L t); auto.
        destruct (ZZ (pt L t)).
         destruct (Z3 L (EBindFac t0)) as (E', H1); try exact I.
         rewrite H1.
         rewrite H1 in IHEval2.
         rewrite eval_val with (v1 := pt L t) (v2 := pt L v1); auto.
        apply E_Ctxt with (E := EBindFac (pt L t0)) (v1 := pt L v1) (t := pt L t); auto.
       simpl; simpl in IHEval2; destruct (L l); auto.
       rewrite <- double_eval with (pt L t) (pt L v1) (pt L v2); auto.
      simpl; simpl in IHEval2; destruct (L l); auto.
      rewrite <- double_eval with (pt L t) (pt L v1) (pt L v2); auto.
     destruct (ZZ (pt L t)).
      destruct (Z3 L EProd) as (E', H1); try exact I.
      rewrite H1.
      rewrite H1 in IHEval2.
      rewrite eval_val with (v1 := pt L t) (v2 := pt L v1); auto.
     apply E_Ctxt with (E := EProd) (v1 := pt L v1) (t := pt L t); auto.
    apply E_Ret.
   apply E_Bind_P; auto.
  simpl; destruct (L k); auto.
 apply E_Bind_B.
Qed.

Definition Equiv (s1 s2 : Store) :=
    (forall a, read s1 a = read s2 a)
  /\
    (forall h, contents s1 h = contents s2 h).

Lemma Z5 : forall pc L k, Visible pc L -> pc k = Some true -> L k = true.
 firstorder.
Qed.
Lemma Z6 : forall pc L k, Visible pc L -> pc k = Some false -> L k = false.
 firstorder.
Qed.

Lemma Z7 : forall pc L k, NotVisible pc L -> pc k = None -> NotVisible (add pc k) L.
 intros.
 unfold NotVisible.
 intro.
 apply H.
 unfold Visible.
 split.
  intro.
  apply H1.
  compute.
  destruct (eq_dec_label k k0); auto.
  intro.
  apply H1.
  compute.
  destruct (eq_dec_label k k0); auto.
 congruence.
Qed.
Lemma Z8 : forall pc L k, NotVisible pc L -> pc k = None -> NotVisible (subtract pc k) L.
 intros.
 unfold NotVisible.
 intro.
 apply H.
 unfold Visible.
 split.
  intro.
  apply H1.
  compute.
  destruct (eq_dec_label k k0); auto.
  congruence.
 intro.
 apply H1.
 compute.
 destruct (eq_dec_label k k0); auto.
Qed.
Variable Z15 : forall s1 s2, Equiv s1 s2 -> s1 = s2.
Lemma Z9 : forall L s a t, ps L (overwrite s a t) = overwrite (ps L s) a (pt L t).
 intros.
 apply Z15.
 unfold Equiv.
 split.
  intros.
  destruct s.
  simpl.
  destruct (eq_dec_addr a a0); auto.
 intros.
 destruct s.
 simpl.
 destruct (eq_dec_view L (view h)); auto.
Qed.

Variable Z10: forall pc L F1 F2, NotVisible pc L -> pt L (facet pc F1 F2) = pt L F2.

Lemma Z11: forall s a t, read (overwrite s a t) a = t.
 intros.
 destruct s.
 simpl.
 destruct eq_dec_addr; intuition.
Qed.
Lemma Z12: forall s a, NotInDomain a s -> read s a = TBottom.
 auto.
Qed.
Lemma Z13 : forall s, Equiv s s.
 split; auto.
Qed.
Lemma Z14 : forall pc L, Visible pc L -> NotVisible pc L -> False.
 auto.
Qed.

Theorem invisibility :
    forall pc L s A s' t,
      NotVisible pc L ->
      Run pc s A s' t ->
      Equiv (ps L s) (ps L s').
 induction 2; auto; try apply Z13.
      destruct IHRun1; auto; destruct IHRun2; auto; split; congruence.
     destruct IHRun1; try apply Z7; auto; destruct IHRun2; try apply Z8; auto; split; congruence.
    rewrite Z9.
    split; auto.
    rewrite Z10; auto.
    simpl; intros.
    destruct (eq_dec_addr a a0) as [H2|H2]; auto.
    rewrite <- H2; rewrite Z12; auto.
   rewrite Z9.
   split; auto.
   simpl; intros.
   destruct (eq_dec_addr a a0) as [H2|H2]; auto.
   rewrite Z10; auto.
   congruence.
  split; destruct s; auto.
  intros; simpl.
  destruct (eq_dec_view L (view h0)); auto.
  destruct (eq_dec_handle h h0); auto.
  destruct Z14 with pc L; congruence.
 split; destruct s; auto.
 intros; simpl.
 destruct (eq_dec_view L (view h0)); auto.
 destruct (eq_dec_handle h h0); auto.
 destruct Z14 with pc L; congruence.
Qed.

Lemma Z16 : forall pc L k, Visible pc L -> L k = true -> Visible (add pc k) L.
 unfold Visible.
 split; intros; unfold add in H1; destruct (eq_dec_label k k0); try (apply H; auto; fail).
  congruence.
 discriminate.
Qed.
Lemma Z17 : forall pc L k, L k = true -> NotVisible (subtract pc k) L.
 unfold NotVisible, not, Visible.
 intros.
 cut (L k = false); [intros; congruence |].
 destruct (H0 k).
 apply H2.
 unfold subtract.
 destruct (eq_dec_label k k); congruence.
Qed.
Lemma Z18 : forall pc L k, Visible pc L -> L k = false -> Visible (subtract pc k) L.
 unfold Visible.
 split; intros; unfold subtract in H1; destruct (eq_dec_label k k0); try (apply H; auto; fail).
  discriminate.
 congruence.
Qed.
Lemma Z19 : forall pc L k, L k = false -> NotVisible (add pc k) L.
 unfold NotVisible, not, Visible.
 intros.
 cut (L k = true); [intros; congruence |].
 destruct (H0 k).
 apply H1.
 unfold add.
 destruct (eq_dec_label k k); congruence.
Qed.
Definition eq_dec_bool : forall (b1 b2 : bool), {b1=b2}+{b1<>b2}.
 decide equality.
Qed.
Lemma Z20 : forall pc L F1 F2, Visible pc L -> pt L (facet pc F1 F2) = pt L F1.
 intros.
 unfold facet.
 induction all_labels; auto.
 remember (pc a).
 destruct o; try destruct b; simpl; remember (L a); destruct b; auto; contradict Heqb.
  assert (true = L a); [| intro; congruence].
  symmetry.
  apply H.
  intuition.
 assert (false = L a); [| intro; congruence].
 symmetry.
 apply H.
 intuition.
Qed.
Lemma Z22 : forall a s L, NotInDomain a s -> NotInDomain a (ps L s).
 intros.
 destruct s.
 unfold ps.
 unfold NotInDomain in *.
 simpl in *.
 rewrite H.
 reflexivity.
Qed.
Variable Z23 : forall s h z, ps (view h) (replace s h z) = replace (ps (view h) s) h z.
Variable Z24 : forall L F1 F2, pt L (facet (view2pc L) F1 F2) = F1.
Lemma Z25 : forall L s h z, L<>view h -> ps L (replace s h z) = ps L s.
 intros.
 apply Z15.
 split.
  intros.
  destruct s.
  intuition.
 intros.
 destruct s.
 simpl.
 destruct (eq_dec_view L (view h0)); auto.
 destruct (eq_dec_handle h h0).
  congruence.
 auto.
Qed.
Variable Z26 : forall L L' F1 F2, L<>L' -> pt L (facet (view2pc L') F1 F2) = F2.
Lemma Z27 : forall s h, contents (ps (view h) s) h = contents s h.
 intros; simpl.
 destruct (eq_dec_view (view h) (view h)); congruence.
Qed.

Theorem impure_projection :
    forall pc s A s' t L,
      Visible pc L ->
      Run pc s A s' t ->
      Run' (ps L s) (pt L A) (ps L s') (pt L t).
 induction 2.
             apply G_Ret.
            apply G_Bind with (pt L t3) (pt L A1) (pt L A2) (ps L s1); [
                apply pure_projection; auto
              |
              | apply pure_projection with (t := TApp t2 t3)
              |
            ]; auto.
           apply G_Prod_P with (A := pt L A); [
               apply pure_projection
             |
           ]; auto.
          assert (L k = true); [apply Z5 with pc; auto |].
          simpl; destruct (L k); try discriminate; auto.
         assert (L k = false); [apply Z6 with pc; auto |].
         simpl; destruct (L k); try discriminate; auto.
        simpl.
        assert (L k = true \/ L k = false); [destruct (L k); auto |].
        destruct H1.
         rewrite H1.
         rewrite <- Z15 with (ps L s1) (ps L s2).
          apply IHRun1.
          apply Z16; auto.
         apply (invisibility (subtract pc k) L s1 (TProd F2) s2 t2); try apply Z17; auto.
        rewrite H1.
        rewrite Z15 with (ps L s0) (ps L s1).
         apply IHRun2.
         apply Z18; auto.
        apply (invisibility (add pc k) L s0 (TProd F1) s1 t1); try apply Z19; auto.
       rewrite Z9.
       rewrite Z20; auto.
       apply G_New; try apply pure_projection; auto.
       apply Z22; auto.
      apply G_Read with (s := ps L s) (t := pt L t).
      rewrite (eq_refl : TAddress a = pt L (TAddress a)).
      apply pure_projection; auto.
     rewrite Z9.
     rewrite Z20; auto.
     apply G_Write with (t1 := pt L t1).
     rewrite (eq_refl : TAddress a = pt L (TAddress a)).
     apply pure_projection; auto.
    destruct (eq_dec_view L (view h)) as [H3|H3].
     rewrite H3.
     rewrite Z23.
     rewrite Z24.
     unfold pt at 1; destruct (eq_dec_view (view h) (view h)); try intuition.
     apply G_Get.
     simpl.
     destruct (eq_dec_view (view h) (view h)); intuition.
    rewrite Z25; auto.
    rewrite Z26; auto.
    unfold pt; destruct (eq_dec_view L (view h)); try congruence.
    apply G_Ret.
   assert (L <> view h); [intro; apply Z14 with pc L; congruence |].
   unfold pt; destruct (eq_dec_view L (view h)); try intuition.
   apply G_Ret.
  simpl.
  destruct (eq_dec_view L (view h)) as [H2|H2].
   rewrite H2.
   rewrite Z23.
   rewrite <- Z27.
   apply G_Put.
   rewrite (eq_refl : TChar ch = pt (view h) (TChar ch)).
   apply pure_projection.
   auto.
  rewrite Z25; auto.
  apply G_Ret.
 assert (L <> view h); [intro; apply Z14 with pc L; congruence |].
 unfold pt; destruct (eq_dec_view L (view h)); try intuition.
 apply G_Ret.
Qed.

Lemma run'_deterministic :
    forall sigma1 A_1 sigma1' t1',
      Run' sigma1 A_1 sigma1' t1' ->
      forall sigma2 A_2 sigma2' t2',
      EQUIV sigma1 A_1 sigma2 A_2 ->
      Run' sigma2 A_2 sigma2' t2' ->
      EQUIV sigma1' t1' sigma2' t2'.
Admitted.

Theorem noninterference : NONINTERFERENCE.
 unfold NONINTERFERENCE.
 intros.
 eapply run'_deterministic.
   apply impure_projection with empty.
    unfold Visible.
    intros.
    split.
     intro.
     discriminate.
    discriminate.
   exact H0.
  exact H.
 apply impure_projection with empty.
  unfold Visible.
  intros.
  split.
   intro.
   discriminate.
  discriminate.
 exact H1.
Qed.

(* END OF PROOF *)
